import os
import re

from setuptools import find_packages, setup
from vexbot.extension_metadata import extensions


VERSIONFILE = 'vexbot/_version.py'
verstrline = open(VERSIONFILE, 'rt').read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in {}".format(VERSIONFILE))


# directory = os.path.abspath(os.path.dirname(__file__))
"""
with open(os.path.join(directory, 'README.rst')) as f:
    long_description = f.read()
"""

_str_form = '{}={}{}'
_extensions = []
for name, extension in extensions.items():
    extras = extension.get('extras')
    if extras is None:
        extras = ''
    # FIXME: This will error out weirdly if there's not a list
    else:
        extras = ', '.join(extras)
        extras = ' [' + extras + ']'
    line = _str_form.format(name, extension['path'], extras)
    _extensions.append(line)

setup(
    name="vexbot",
    version=verstr,
    description='Python personal assistant',
    # long_description=long_description,
    url='https://github.com/benhoff/vexbot',
    license='GPL3',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent'],
    author='Ben Hoff',
    author_email='beohoff@gmail.com',
    entry_points={'console_scripts': ['vexbot=vexbot.adapters.shell.__main__:main',
                                      'vexbot_robot=vexbot.__main__:main',
                                      'vexbot_irc=vexbot.adapters.irc.__main__:main',
                                      'vexbot_xmpp=vexbot.adapters.xmpp:main',
                                      'vexbot_socket_io=vexbot.adapters.socket_io.__main__:main',
                                      'vexbot_youtube=vexbot.adapters.youtube:main',
                                      'vexbot_stackoverflow=vexbot.adapters.stackoverflow:main',
                                      'vexbot_generate_certificates=vexbot.util.generate_certificates:main',
                                      'vexbot_generate_unit_file=vexbot.util.generate_config_file:main'],
                    'vexbot_extensions': _extensions},
    packages=find_packages(), # exclude=['docs', 'tests']
    dependency_links=[
        'git+https://github.com/jonathanslenders/python-prompt-toolkit@2.0',
	'git+https://github.com/benhoff/vexmessage@dev'
        ],

    install_requires=[
        # 'pluginmanager>=0.4.1',
        'pyzmq',
        'vexmessage>=0.4.0',
        'rx',
        'tblib', # traceback serilization
        'tornado',
        ],

    extras_require={
        'nlp': ['wheel', 'spacy', 'sklearn', 'sklearn_crfsuite'],
        'socket_io': ['requests', 'websocket-client'],
        'summarization': ['gensim', 'newspaper3k'],
        'youtube': ['google-api-python-client'],
        'dev': ['flake8', 'twine', 'wheel'],
        'xmpp': ['sleekxmpp', 'dnspython'],
        'process_name': ['setproctitle'],
        'speechtotext': ['speechtotext'],
        'digitalocean': ['python-digitalocean'],
        'process_manager': ['pydbus'],
        'microphone': ['microphone'],
        'database': ['vexstorage'],
        'gui': ['chatimusmaximus'],
        'entity': ['duckling'],
        'irc': ['irc3'],
        'system': ['psutil'],
        'command_line': ['python-prompt-toolkit', 'pygments'],
    }
)



import argparse as _argparse


def get_kwargs():
    parser = _argparse.ArgumentParser()
    parser.add_argument('--configuration_filepath', default=None)
    args = parser.parse_args()
    return vars(args)



import unittest
from time import sleep

import zmq
from vexmessage import decode_vex_message
from vexbot.messaging import Messaging


class TestMessaging(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        self.subscribe_address = 'tcp://127.0.0.1:4006'
        self.publish_address = 'tcp://127.0.0.1:4007'
        self.settings = {'subscribe_address': self.subscribe_address,
                         'publish_address': self.publish_address}

        context = zmq.Context()
        self.messaging = Messaging(self.settings, context)

        self.test_publish_socket = context.socket(zmq.PUB)
        self.test_publish_socket.connect(self.publish_address)

        self.test_subscribe_socket = context.socket(zmq.SUB)
        self.test_subscribe_socket.connect(self.subscribe_address)
        self.test_subscribe_socket.setsockopt_string(zmq.SUBSCRIBE, '')
        sleep(.2)
        super().__init__(*args, **kwargs)

    """
    def test_send_message(self):
        self.messaging.send_message('', test='blue')
        frame = self.test_subscribe_socket.recv_multipart()
        message = decode_vex_message(frame)
        self.assertEqual(message.target, '')
        self.assertEqual(message.source, 'robot')
        self.assertEqual(message.contents['test'], 'blue')
        self.assertEqual(message.type, 'MSG')
    """

    def test_send_targeted_message(self):
        self.test_subscribe_socket.setsockopt_string(zmq.SUBSCRIBE, 'test')
        self.messaging.send_message('test', test='blue')
        frame = self.test_subscribe_socket.recv_multipart()
        message = decode_vex_message(frame)
        self.assertEqual(message.target, 'test')
        self.assertEqual(message.source, 'robot')
        self.assertEqual(message.contents['test'], 'blue')
        self.assertEqual(message.type, 'MSG')



from zmq.eventloop.zmqstream import ZMQStream


class Scheduler:
    def __init__(self):
        # FIXME: Move into the SocketFactory
        self._streams = []

    def register_socket(self, socket, loop, on_recv):
        stream = ZMQStream(socket, loop)
        stream.on_recv(on_recv)
        self._streams.append(stream)



import argparse
import json
import html
import logging
import pkg_resources
import atexit
from time import sleep
from threading import Thread

from vexmessage import decode_vex_message

from vexbot.adapters.messaging import Messaging 
from vexbot.adapters.socket_io.observer import SocketObserver
try:
    from websocket import WebSocketApp
except ImportError:
    logging.error('Socket IO needs `websocket` installed. Please run `pip '
                  'install websocket-client`')
try:
    import requests
except ImportError as e:
    logging.error('Socket IO needs `requests` installed. Please run `pip '
                  'install requests`')


class WebSocket(WebSocketApp):
    def __init__(self,
                 streamer_name: str,
                 namespace: str,
                 website_url: str,
                 service_name: str,
                 connection: dict=None):

        self.log = logging.getLogger(__name__)
        self.log.setLevel(0)
        if connection is None:
            connection = {}

        self.messaging = Messaging(service_name, run_control_loop=True, **connection)
        self._scheduler_thread = Thread(target=self.messaging.start,
                                        daemon=True)

        self._scheduler_thread.start()
        self.observer = SocketObserver(self, self.messaging)
        self.messaging.command.subscribe(self.observer)

        self._streamer_name = streamer_name
        self.namespace = namespace
        self._website_url = website_url
        self.log.info('Getting Socket IO key!')
        self.key, heartbeat = self._connect_to_server_helper()
        self.log.info('Socket IO key got!')
        # self.command_manager = AdapterCommandManager(self.messaging)

        # alters URL to be more websocket...ie
        self._website_socket = self._website_url.replace('http', 'ws')
        self._website_socket += 'websocket/'
        self.nick = None
        super().__init__(self._website_socket + self.key,
                         on_open=self.on_open,
                         on_close=self.on_close,
                         on_message=self.on_message,
                         on_error=self.on_error)

    def _connect_to_server_helper(self):
        r = requests.post(self._website_url)
        params = r.text

        # unused variables are connection_timeout and supported_formats
        key, heartbeat_timeout, _, _ = params.split(':')
        heartbeat_timeout = int(heartbeat_timeout)
        return key, heartbeat_timeout

    def on_open(self, *args):
        logging.info('Websocket open!')

    def on_close(self, *args):
        logging.info('Websocket closed :(')

    def on_message(self, *args):
        message = args[1].split(':', 3)
        key = int(message[0])
        # namespace = message[2]

        if len(message) >= 4:
            data = message[3]
        else:
            data = ''
        if key == 1 and args[1] == '1::':
            self.send_packet_helper(1)
        elif key == 1 and args[1] == '1::{}'.format(self.namespace):
            self.send_packet_helper(5, data={'name': 'initialize'})
            data = {'name': 'join',
                    'args': ['{}'.format(self._streamer_name)]}

            self.send_packet_helper(5, data=data)
            self.log.info('Connected to channel with socket io!')
            # self.messaging.send_status('CONNECTED')
        elif key == 2:
            self.send_packet_helper(2)
        elif key == 5:
            data = json.loads(data, )
            if data['name'] == 'message':
                message = data['args'][0]
                sender = html.unescape(message['sender'])
                if sender == self.nick:
                    return
                message = html.unescape(message['text'])
                self.messaging.send_chatter(author=sender, message=message)
            elif data['name'] == 'join':
                self.nick = data['args'][1]

    def on_error(self, *args):
        logging.error(args[1])

    def disconnect(self):
        callback = ''
        data = ''
        # '1::namespace'
        self.send(':'.join([str(self.TYPE_KEYS['DISCONNECT']),
                           callback, self.namespace, data]))

    def send_packet_helper(self,
                           type_key,
                           data=None):

        if data is None:
            data = ''
        else:
            data = json.dumps(data)

        # NOTE: callbacks currently not implemented
        callback = ''
        message = ':'.join([str(type_key), callback, self.namespace, data])
        self.send(message)



import sys as _sys
import zmq as _zmq
import logging
from os import path

from zmq.auth.thread import ThreadAuthenticator
from zmq.auth.ioloop import IOLoopAuthenticator

from vexbot.util.get_certificate_filepath import (get_certificate_filepath,
                                                  get_vexbot_certificate_filepath,
                                                  get_client_certificate_filepath)

class SocketFactory:
    """
    Abstracts out the socket creation, specifically the issues with
    transforming ports into zmq addresses.
    Also sets up some default protocol and address handeling.
    For example, if everything is being run locally on one machine, and the
    default is to bind all addresses `*` and the transport protocol
    `tcp` means tcp communication
    """
    def __init__(self,
                 address: str,
                 protocol: str='tcp',
                 context: 'zmq.Context'=None,
                 logger: 'logging.Logger'=None,
                 loop=None,
                 auth_whitelist: list=None,
                 using_auth: bool=True):

        self.address = address
        self.protocol = protocol
        self.context = context or _zmq.Context.instance()
        if logger is None:
            logger = logging.getLogger(__name__)
        self.logger = logger
        self._server_certs = (None, None)
        self.using_auth = using_auth 

        if self.using_auth:
            self._setup_auth(auth_whitelist, loop)

    def _setup_auth(self, auth_whitelist: list, loop):
        if loop is None:
            self.auth = ThreadAuthenticator(self.context)
        else:
            self.auth = IOLoopAuthenticator(self.context, io_loop=loop)

        # allow all local host
        self.logger.debug('Auth whitelist %s', auth_whitelist)
        if auth_whitelist is not None:
            self.auth.allow(auth_whitelist)
        self._base_filepath = get_certificate_filepath()
        public_key_location = path.join(self._base_filepath, 'public_keys')
        self.auth.configure_curve(domain='*', location=public_key_location)
        self.auth.start()

    def _set_server_certs(self, any_=True):
        secret_filepath, secret = get_vexbot_certificate_filepath()
        if not secret and not any_:
            err = ('Could not find the vexbot certificates. Generate them '
                   'using `vexbot_generate_certificates` from command line')
            raise FileNotFoundError(err)

        self._server_certs = _zmq.auth.load_certificate(secret_filepath)

    def _create_using_auth(self, socket, address, bind, on_error, socket_name):
        if not any(self._server_certs):
            self._set_server_certs(bind)
        if bind:
            if self._server_certs[1] is None:
                err = ('Server Secret File Not Found! Generate using '
                       '`vexbot_generate_certificates` from the command line')
                raise FileNotFoundError(err)
                                        
            public, private = self._server_certs
        else:
            # NOTE: This raises a `FileNotFoundError` if not found
            secret_filepath = get_client_certificate_filepath()
            public, private = _zmq.auth.load_certificate(secret_filepath)

        socket.curve_secretkey = private
        socket.curve_publickey = public

        if bind:
            socket.curve_server = True
            try:
                socket.bind(address)
            except _zmq.error.ZMQError:
                self._handle_error(on_error, address, socket_name)
                socket = None
        else:  # connect the socket
            socket.curve_serverkey = self._server_certs[0]
            socket.connect(address)

    def _create_no_auth(self, socket, address, bind, on_error, socket_name):
        if bind:
            try:
                socket.bind(address)
            except _zmq.error.ZMQError:
                self._handle_error(on_error, address, socket_name)
                socket = None
        else:  # connect the socket
            socket.connect(address)

    def create_n_connect(self,
                         socket_type,
                         address: str,
                         bind=False,
                         on_error='log',
                         socket_name=''):
        """
        Creates and connects or binds the sockets
        on_error:
            'log': will log error
            'exit': will exit the program
        socket_name:
            used for troubleshooting/logging
        """
        self.logger.debug('create and connect: %s %s %s',
                          socket_type, socket_name, address)

        socket = self.context.socket(socket_type)
        if self.using_auth:
            self._create_using_auth(socket, address, bind, on_error, socket_name)
        else:
            self._create_no_auth(socket, address, bind, on_error, socket_name)

        return socket

    def multiple_create_n_connect(self,
                                  socket_type,
                                  addresses: tuple,
                                  bind=False,
                                  on_error='log',
                                  socket_name=''):

        # TODO: Refactor this to remove code duplication with create_n_connect
        # Or at least make better sense
        socket = self.context.socket(socket_type)
        for address in addresses:
            if self.auth:
                self._create_using_auth(socket, address, bind, on_error, socket_name)
            else:
                self._create_no_auth(socket, address, bind, on_error, socket_name)
        return socket

    def to_address(self, port: str):
        """
        transforms the ports into addresses.
        Will fall through if the port looks like an address
        """
        # check to see if user passed in a string
        # if they did, they want to use that instead
        if isinstance(port, str) and len(port) > 6:
            return port

        zmq_address = '{}://{}:{}'
        return zmq_address.format(self.protocol,
                                  self.address,
                                  port)

    def iterate_multiple_addresses(self, ports: (str, list, tuple)):
        """
        transforms an iterable, or the expectation of an iterable
        into zmq addresses
        """
        if isinstance(ports, (str, int)):
            # TODO: verify this works.
            ports = tuple(ports,)

        result = []
        for port in ports:
            result.append(self.to_address(port))

        return result

    def _handle_error(self, how_to: str, address: str, socket_name: str):
        if socket_name == '':
            socket_name = 'unknown type'

        if how_to == 'exit':
            self._handle_bind_error_by_exit(address, socket_name)
        else:
            self._handle_bind_error_by_log(address, socket_name)

    def _handle_bind_error_by_log(self, address: str, socket_name: str):
        if self.logger is None:
            return
        s = 'Address bind attempt fail. Address tried: {}'
        s = s.format(address)
        self.logger.error(s)
        self.logger.error('socket type: {}'.format(socket_name))

    def _handle_bind_error_by_exit(self, address: str, socket_type: str):
        if self.logger is not None:
            s = 'Address bind attempt fail. Alredy in use? Address tried: {}'
            s = s.format(address)
            self.logger.error(s)
            self.logger.error('socket type: {}'.format(socket_type))

        _sys.exit(1)



import inspect as _inspect

from vexmessage import Request
from vexbot.observer import Observer
from vexbot.extensions import develop
from vexbot.extensions import help as vexhelp


class SocketObserver(Observer):
    extensions = (develop.get_code, develop.get_commands, vexhelp.help)
    def __init__(self, bot, messaging):
        super().__init__()
        self.bot = bot
        self.messaging = messaging
        self._commands = self._get_commands()

    def _get_commands(self) -> dict:
        result = {}
        for name, method in _inspect.getmembers(self):
            if name.startswith('do_'):
                result[name[3:]] = method

        return result

    def do_MSG(self, message, *args, **kwargs):
        msg_target = kwargs.get('msg_target')
        if msg_target:
            message = msg_target + ', ' + message

        data = {}
        data['name'] = 'message'
        data['args'] = [message, self.bot._streamer_name]

        # FIXME: this API is awful.
        self.bot.send_packet_helper(5, data)

    def on_next(self, item: Request):
        command = item.command
        args = item.args
        kwargs = item.kwargs
        try:
            callback = self._commands[command]
        except KeyError:
            return

        try:
            result = callback(*args, **kwargs)
        except Exception as e:
            self.on_error(e, command, args)
            return

        if result is None:
            return

        source = item.source
        self.messaging.send_command_response(source, command, result=result, *args, **kwargs)

    def on_error(self, *args, **kwargs):
        pass

    def on_completed(self, *args, **kwargs):
        pass
